# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dimod.core.composite import ComposedSampler
from dimod.sampleset import SampleSet

__all__ = ['ClipComposite']

class ClipComposite(ComposedSampler):
    """Composite to clip variables of a problem.

    Clips the variables of a binary quadratic model (BQM) and modifies linear
    and quadratic terms accordingly.

    Args:
       sampler (:class:`dimod.Sampler`):
            A dimod sampler.

    Examples:
       This example uses :class:`.ClipComposite` to instantiate a
       composed sampler that submits a simple Ising problem to a sampler.
       The composed sampler clips linear and quadratic biases as
       indicated by options.

       >>> from dimod import ExactSolver
       >>> from dwave.preprocessing.composites import ClipComposite
       >>> h = {'a': -4.0, 'b': -4.0}
       >>> J = {('a', 'b'): 3.2}
       >>> sampler = ClipComposite(ExactSolver())
       >>> response = sampler.sample_ising(h, J, lower_bound=-2.0, upper_bound=2.0)

    """

    def __init__(self, child_sampler):
        self._children = [child_sampler]

    @property
    def children(self):
        return self._children

    @property
    def parameters(self):
        param = self.child.parameters.copy()
        param.update({'lower_bound': [], 'upper_bound': []})
        return param

    @property
    def properties(self):
        return {'child_properties': self.child.properties.copy()}

    def sample(self, bqm, *, lower_bound=None, upper_bound=None, **parameters):
        """Clip and sample from the provided binary quadratic model.

        If lower_bound and upper_bound are given variables with value above or below are clipped.

        Args:
            bqm (:class:`dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.

            lower_bound (number):
                Value by which to clip the variables from below.

            upper_bound (number):
                Value by which to clip the variables from above.

            **parameters:
                Parameters for the sampling method, specified by the child sampler.

        Returns:
            :class:`dimod.SampleSet`

        """
        child = self.child
        bqm_copy = _clip_bqm(bqm, lower_bound, upper_bound)
        response = child.sample(bqm_copy, **parameters)

        return SampleSet.from_samples_bqm(response, bqm, info=response.info)


def _clip_bqm(bqm, lower_bound, upper_bound):
    """Helper function for clipping a bqm."""

    bqm_copy = bqm.copy()
    if lower_bound is not None:
        linear = bqm_copy.linear
        for k, v in linear.items():
            if v < lower_bound:
                linear[k] = lower_bound
        quadratic = bqm_copy.quadratic
        for k, v in quadratic.items():
            if v < lower_bound:
                quadratic[k] = lower_bound

    if upper_bound is not None:
        linear = bqm_copy.linear
        for k, v in linear.items():
            if v > upper_bound:
                linear[k] = upper_bound
        quadratic = bqm_copy.quadratic
        for k, v in quadratic.items():
            if v > upper_bound:
                quadratic[k] = upper_bound
    return bqm_copy



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dimod.core.composite import ComposedSampler
from dimod.decorators import nonblocking_sample_method
from dimod.sampleset import SampleSet

__all__ = ['ScaleComposite']

class ScaleComposite(ComposedSampler):
    """Composite that scales variables of a problem.

    Scales the variables of a binary quadratic model (BQM) and modifies linear
    and quadratic terms accordingly.

    Args:
       sampler (:class:`dimod.Sampler`):
            A dimod sampler.

    Examples:
       This example uses :class:`.ScaleComposite` to instantiate a
       composed sampler that submits a simple Ising problem to a sampler.
       The composed sampler scales linear biases, quadratic biases, and
       offset as indicated by options.

       >>> from dimod import ExactSolver
       >>> from dwave.preprocessing.composites import ScaleComposite
       >>> h = {'a': -4.0, 'b': -4.0}
       >>> J = {('a', 'b'): 3.2}
       >>> sampler = ScaleComposite(ExactSolver())
       >>> response = sampler.sample_ising(h, J, scalar=0.5,
       ...                ignored_interactions=[('a','b')])

    """

    def __init__(self, child_sampler):
        self._children = [child_sampler]

    @property
    def children(self):
        return self._children

    @property
    def parameters(self):
        param = self.child.parameters.copy()
        param.update({'scalar': [],
                      'bias_range': [],
                      'quadratic_range': [],
                      'ignored_variables': [],
                      'ignored_interactions': [],
                      'ignore_offset': []})
        return param

    @property
    def properties(self):
        return {'child_properties': self.child.properties.copy()}

    @nonblocking_sample_method
    def sample(self, bqm, *, scalar=None, bias_range=1, quadratic_range=None,
               ignored_variables=None, ignored_interactions=None,
               ignore_offset=False, **parameters):
        """Scale and sample from the provided binary quadratic model.

        If ``scalar`` is not given, the problem is scaled based on bias and 
        quadratic ranges. See :meth:`.BinaryQuadraticModel.scale` and
        :meth:`.BinaryQuadraticModel.normalize`

        Args:
            bqm (:class:`dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.

            scalar (number):
                Value by which to scale the energy range of the binary
                quadratic model. Overrides `bias_range` and `quadratic_range`.

            bias_range (number/pair, default=1):
                Value/range by which to normalize the all the biases, or if
                `quadratic_range` is provided, just the linear biases.
                Overridden by `scalar`.

            quadratic_range (number/pair):
                Value/range by which to normalize the quadratic biases.
                Overridden by `scalar`.

            ignored_variables (iterable, optional):
                Biases associated with these variables are not scaled.

            ignored_interactions (iterable[tuple], optional):
                As an iterable of 2-tuples. Biases associated with these
                interactions are not scaled.

            ignore_offset (bool, default=False):
                If True, the offset is not scaled.

            **parameters:
                Parameters for the sampling method, specified by the child
                sampler.

        Returns:
            :class:`dimod.SampleSet`

        """
        original_bqm = bqm
        bqm = bqm.copy()  # we're going to be scaling

        if scalar is not None:
            bqm.scale(scalar,
                      ignored_variables=ignored_variables,
                      ignored_interactions=ignored_interactions,
                      ignore_offset=ignore_offset)
        else:
            scalar = bqm.normalize(bias_range, quadratic_range,
                                   ignored_variables=ignored_variables,
                                   ignored_interactions=ignored_interactions,
                                   ignore_offset=ignore_offset)

        if scalar == 0:
            raise ValueError('scalar must be non-zero')

        sampleset = self.child.sample(bqm, **parameters)

        yield sampleset  # so that SampleSet.done() works

        if not (ignored_variables or ignored_interactions or ignore_offset):
            # we just need to scale back and don't need to worry about
            # the stuff we ignored
            sampleset.record.energy *= 1 / scalar
        else:
            sampleset.record.energy = original_bqm.energies(sampleset)

        sampleset.info.update(scalar=scalar)

        yield sampleset



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dimod.vartypes import Vartype

from dwave.preprocessing.cyfix_variables import fix_variables_wrapper

def roof_duality(bqm, *, strict=True):
    """Determine minimizing assignments for some variables of a binary quadratic 
    model using the roof duality algorithm.

    Args:
        bqm (:class:`.BinaryQuadraticModel`):
            A binary quadratic model.

        strict (bool, optional, default=True):
            If True, only fixes variables for which assignments are true for all 
            minimizing points (strong persistency). If False, also fixes variables 
            for which the assignments are true for some but not all minimizing 
            points (weak persistency).

    Returns:
        dict: Variable assignments for some variables of ``bqm``.

    Examples:
        This example creates a binary quadratic model with a single ground state
        and fixes the model's single variable to the minimizing assignment.

        >>> import dimod
        >>> from dwave.preprocessing.lower_bounds import roof_duality
        >>> bqm = dimod.BinaryQuadraticModel.from_ising({'a': 1.0}, {})
        >>> roof_duality(bqm)
        {'a': -1}

        This example has two ground states, :math:`a=b=-1` and :math:`a=b=1`, with
        no variable having a single value for all ground states, so neither variable
        is fixed.

        >>> bqm = dimod.BinaryQuadraticModel.empty(dimod.SPIN)
        >>> bqm.add_interaction('a', 'b', -1.0)
        >>> roof_duality(bqm) # doctest: +SKIP
        {}

        This example sets ``strict`` to False, so variables are fixed to an assignment
        that attains the ground state.

        >>> bqm = dimod.BinaryQuadraticModel.empty(dimod.SPIN)
        >>> bqm.add_interaction('a', 'b', -1.0)
        >>> roof_duality(bqm, strict=False) # doctest: +SKIP
        {'a': 1, 'b': 1}

    """
    bqm_ = bqm

    if bqm_.vartype is Vartype.SPIN:
        bqm_ = bqm.change_vartype(Vartype.BINARY, inplace=False)

    linear = bqm_.linear

    if all(v in linear for v in range(len(bqm_))):
        # we can work with the binary form of the bqm directly
        fixed = fix_variables_wrapper(bqm_, strict)
    else:
        inverse_mapping = dict(enumerate(linear))
        mapping = {v: i for i, v in inverse_mapping.items()}

        # no need to make a copy if we've already done so
        inplace = bqm.vartype is Vartype.SPIN
        bqm_ = bqm_.relabel_variables(mapping, inplace=inplace)

        fixed = fix_variables_wrapper(bqm_, strict)
        fixed = {inverse_mapping[v]: val for v, val in fixed.items()}

    if bqm.vartype is Vartype.SPIN:
        return {v: 2*val - 1 for v, val in fixed.items()}
    else:
        return fixed



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from random import random

from dimod.core.composite import Composite
from dimod.core.sampler import Sampler
from dimod.sampleset import SampleSet, concatenate
from dimod.vartypes import Vartype

__all__ = ['SpinReversalTransformComposite']

class SpinReversalTransformComposite(Sampler, Composite):
    """Composite for applying spin reversal transform preprocessing.

    Spin reversal transforms (or "gauge transformations") are applied
    by flipping the spin of variables in the Ising problem. After
    sampling the transformed Ising problem, the same bits are flipped in the
    resulting sample [#km]_.

    Args:
        sampler: A `dimod` sampler object.

    Examples:
        This example composes a dimod ExactSolver sampler with spin transforms then
        uses it to sample an Ising problem.

        >>> from dimod import ExactSolver
        >>> from dwave.preprocessing.composites import SpinReversalTransformComposite
        >>> base_sampler = ExactSolver()
        >>> composed_sampler = SpinReversalTransformComposite(base_sampler)
        ... # Sample an Ising problem
        >>> response = composed_sampler.sample_ising({'a': -0.5, 'b': 1.0}, {('a', 'b'): -1})
        >>> response.first.sample
        {'a': -1, 'b': -1}

    References
    ----------
    .. [#km] Andrew D. King and Catherine C. McGeoch. Algorithm engineering
        for a quantum annealing platform. https://arxiv.org/abs/1410.2628,
        2014.

    """
    children = None
    parameters = None
    properties = None

    def __init__(self, child):
        self.children = [child]

        self.parameters = parameters = {'spin_reversal_variables': []}
        parameters.update(child.parameters)

        self.properties = {'child_properties': child.properties}

    def sample(self, bqm, *, num_spin_reversal_transforms=2, **kwargs):
        """Sample from the binary quadratic model.

        Args:
            bqm (:class:`~dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.

            num_spin_reversal_transforms (integer, optional, default=2):
                Number of spin reversal transform runs.

        Returns:
            :class:`.SampleSet`

        Examples:
            This example runs 100 spin reversals applied to one variable of a QUBO problem.

            >>> from dimod import ExactSolver
            >>> from dwave.preprocessing.composites import SpinReversalTransformComposite
            >>> base_sampler = ExactSolver()
            >>> composed_sampler = SpinReversalTransformComposite(base_sampler)
            ...
            >>> Q = {('a', 'a'): -1, ('b', 'b'): -1, ('a', 'b'): 2}
            >>> response = composed_sampler.sample_qubo(Q,
            ...               num_spin_reversal_transforms=100)
            >>> len(response)
            400
        """

        # make a main response
        responses = []

        flipped_bqm = bqm.copy()
        transform = {v: False for v in bqm.variables}

        for ii in range(num_spin_reversal_transforms):
            # flip each variable with a 50% chance
            for v in bqm:
                if random() > .5:
                    transform[v] = not transform[v]
                    flipped_bqm.flip_variable(v)

            flipped_response = self.child.sample(flipped_bqm, **kwargs)

            tf_idxs = [flipped_response.variables.index(v)
                       for v, flip in transform.items() if flip]

            if bqm.vartype is Vartype.SPIN:
                flipped_response.record.sample[:, tf_idxs] = -1 * flipped_response.record.sample[:, tf_idxs]
            else:
                flipped_response.record.sample[:, tf_idxs] = 1 - flipped_response.record.sample[:, tf_idxs]

            responses.append(flipped_response)

        return concatenate(responses)



# distutils: language = c++
# cython: language_level=3
#
# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcpp.utility cimport pair
from libcpp.vector cimport vector
from libcpp cimport bool as cppbool

import dimod
from dimod import AdjVectorBQM
from dimod.bqm.cppbqm cimport AdjVectorBQM as cppAdjVectorBQM
from dimod cimport cyAdjVectorBQM
from dimod.vartypes import Vartype

cdef extern from "include/dwave-preprocessing/fix_variables.hpp" namespace "fix_variables_":
    vector[pair[int, int]] fixQuboVariables[V, B](cppAdjVectorBQM[V, B]& refBQM,
                                                  cppbool strict) except +

def fix_variables_wrapper(bqm, strict):
    """Cython wrapper for fix_variables().

    Args:
        bqm (:class:`.BinaryQuadraticModel`):
            A binary quadratic model with binary-valued variables, indexed 
            linearly from zero.

        strict (bool):
            If True, only fixes variables for which assignments are true for all 
            minimizing points (strong persistency). If False, also fixes variables 
            for which the assignments are true for some but not all minimizing 
            points (weak persistency).
    """
    if bqm.vartype is not Vartype.BINARY:
        raise ValueError("bqm must be BINARY")
    if not all(v in bqm.linear for v in range(len(bqm))):
        raise ValueError("bqm must be linearly indexed")

    cdef cyAdjVectorBQM cybqm = dimod.as_bqm(bqm, cls=AdjVectorBQM)
    fixed = fixQuboVariables(cybqm.bqm_, bool(strict))
    return {int(v): int(val) for v, val in fixed}



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from setuptools import setup

import numpy
import dimod

from Cython.Build import cythonize
from distutils.extension import Extension
from distutils.command.build_ext import build_ext as _build_ext

extra_compile_args = {
    'msvc': ['/EHsc'],
    'unix': ['-std=c++11'],
}

extra_link_args = {
    'msvc': [],
    'unix': ['-std=c++11'],
}


class build_ext(_build_ext):
    def build_extensions(self):
        compiler = self.compiler.compiler_type

        compile_args = extra_compile_args[compiler]
        for ext in self.extensions:
            ext.extra_compile_args.extend(compile_args)

        link_args = extra_link_args[compiler]
        for ext in self.extensions:
            ext.extra_compile_args.extend(link_args)

        super().build_extensions()

setup(
    name='dwave-preprocessing',
    cmdclass=dict(build_ext=build_ext),
    ext_modules=cythonize(
        ['dwave/preprocessing/cyfix_variables.pyx'],
        annotate=bool(os.getenv('CYTHON_ANNOTATE', False)),
        nthreads=int(os.getenv('CYTHON_NTHREADS', 0)),
        ),
    include_dirs=[
        numpy.get_include(),
        dimod.get_include(),
        ],
    install_requires=[
        'numpy>=1.17.3,<2.0.0',  # keep synced with circle-ci, pyproject.toml
        'dimod>=0.9.15,<0.10.0'
        ],
)



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import warnings

from dimod.bqm import as_bqm, AdjVectorBQM, AdjDictBQM
from dimod.core.composite import ComposedSampler
from dimod.sampleset import SampleSet, append_variables

from dwave.preprocessing.lower_bounds import roof_duality

__all__ = ['FixVariablesComposite']

class FixVariablesComposite(ComposedSampler):
    """Composite to fix variables of a problem to provided.

    Fixes variables of a binary quadratic model (BQM) and modifies linear and
    quadratic terms accordingly. Returned samples include the fixed variable.

    Args:
        child_sampler (:class:`dimod.Sampler`):
            A dimod sampler
        
        algorithm (str, optional, default='explicit'):
            Determines how ``fixed_variables`` are found. 

            'explicit': ``fixed_variables`` should be passed in a call to 
            `.sample()`. If not, no fixing occurs and the problem is directly 
            passed to the child sampler.

            'roof_duality': Roof duality algorithm is used to find ``fixed_variables``. 
            ``strict`` may be passed in a call to `.sample()` to determine what
            variables the algorithm will fix. For details, see 
            :func:`~dwave.preprocessing.lower_bounds.roof_duality`.

    Examples:
       This example uses the :class:`.FixVariablesComposite` to instantiate a
       composed sampler that submits a simple Ising problem to a sampler.
       The composed sampler fixes a variable and modifies linear and quadratic
       biases accordingly.

       >>> from dimod import ExactSolver
       >>> from dwave.preprocessing.composites import FixVariablesComposite
       >>> h = {1: -1.3, 4: -0.5}
       >>> J = {(1, 4): -0.6}
       >>> sampler = FixVariablesComposite(ExactSolver())
       >>> sampleset = sampler.sample_ising(h, J, fixed_variables={1: -1})

       This next example involves the same problem but calculates ``fixed_variables``
       using the 'roof_duality' ``algorithm``.

       >>> sampler = FixVariablesComposite(ExactSolver(), algorithm='roof_duality')
       >>> sampleset = sampler.sample_ising(h, J, strict=False)

    """

    def __init__(self, child_sampler, *, algorithm='explicit'):
        self._children = [child_sampler]
        self.algorithm = algorithm

        self._parameters = self.child.parameters.copy()

        if self.algorithm == 'explicit':
            self._parameters['fixed_variables'] = []
        elif self.algorithm == 'roof_duality':
            self._parameters['strict'] = []
        else:
            raise ValueError("Unknown algorithm: {}".format(algorithm))

    @property
    def children(self):
        return self._children

    @property
    def parameters(self):
        return self._parameters

    @property
    def properties(self):
        return {'child_properties': self.child.properties.copy()}

    def sample(self, bqm, **parameters):
        """Sample from the provided binary quadratic model.

        Args:
            bqm (:class:`dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.

            fixed_variables (dict, optional, default=None):
                A dictionary of variable assignments used when ``self.algorithm`` 
                is 'explicit'.

            strict (bool, optional, default=True):
                Only used if ``self.algorithm`` is 'roof_duality'. If True, only 
                fixes variables for which assignments are true for all minimizing 
                points (strong persistency). If False, also fixes variables for 
                which the assignments are true for some but not all minimizing 
                points (weak persistency).

            **parameters:
                Parameters for the sampling method, specified by the child sampler.

        Returns:
            :class:`dimod.SampleSet`

        """

        if self.algorithm == 'explicit':
            fixed_variables = parameters.pop('fixed_variables', None)
            if fixed_variables is None:
                msg = ("No fixed_variables passed in when algorithm is 'explicit'. "
                       "Passing problem to child sampler without fixing.")
                warnings.warn(msg)
                return self.child.sample(bqm, **parameters)
        elif self.algorithm == 'roof_duality':
            fixed_variables = roof_duality(bqm, strict=parameters.pop('strict', True))

        # make sure that we're shapeable and that we have a BQM we can mutate
        bqm_copy = as_bqm(bqm, cls=[AdjVectorBQM, AdjDictBQM], copy=True)

        bqm_copy.fix_variables(fixed_variables)

        sampleset = self.child.sample(bqm_copy, **parameters)

        def _hook(sampleset):
            # make RoofDualityComposite non-blocking

            if sampleset.variables:
                if len(sampleset):
                    return append_variables(sampleset, fixed_variables)
                else:
                    return sampleset.from_samples_bqm((np.empty((0, len(bqm))),
                                                       bqm.variables), bqm=bqm)

            # there are only fixed variables, make sure that the correct number
            # of samples are returned
            samples = [fixed_variables]*max(len(sampleset), 1)

            return sampleset.from_samples_bqm(samples, bqm=bqm)

        return SampleSet.from_future(sampleset, _hook)



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dimod.bqm import as_bqm, AdjVectorBQM, AdjDictBQM
from dimod.core.composite import ComposedSampler
from dimod.sampleset import SampleSet, append_variables
from dimod.traversal import connected_components

__all__ = ['ConnectedComponentsComposite']

class ConnectedComponentsComposite(ComposedSampler):
    """Composite to decompose a problem to the connected components
    and solve each.

    Connected components of a binary quadratic model (BQM) graph are computed
    (if not provided), and each subproblem is passed to the child sampler.
    Returned samples from each child sampler are merged. Only the best solution
    of each response is selected and merged with others (i.e. this composite 
    returns a single solution).

    Args:
       sampler (:class:`dimod.Sampler`):
            A dimod sampler

    Examples:
       This example uses :class:`.ConnectedComponentsComposite` to solve a simple
       Ising problem that can be separated into two components. This small example
       uses :class:`dimod.ExactSolver` and is just illustrative.

       >>> from dimod import ExactSolver
       >>> from dwave.preprocessing.composites import ConnectedComponentsComposite
       >>> h = {}
       >>> J1 = {(1, 2): -1.0, (2, 3): 2.0, (3, 4): 3.0}
       >>> J2 = {(12, 13): 6}
       >>> sampler = ExactSolver()
       >>> sampler_ccc = ConnectedComponentsComposite(sampler)
       >>> e1 = sampler.sample_ising(h, J1).first.energy
       >>> e2 = sampler.sample_ising(h, J2).first.energy
       >>> e_ccc = sampler_ccc.sample_ising(h, {**J1, **J2}).first.energy
       >>> e_ccc == e1 + e2
       True

    """

    def __init__(self, child_sampler):
        self._children = [child_sampler]

    @property
    def children(self):
        return self._children

    @property
    def parameters(self):
        params = self.child.parameters.copy()
        return params

    @property
    def properties(self):
        return {'child_properties': self.child.properties.copy()}

    def sample(self, bqm, *, components=None, **parameters):
        """Sample from the provided binary quadratic model.

        Args:
            bqm (:class:`dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.

            components (list(set)):
                A list of disjoint set of variables that fully partition the variables

            **parameters:
                Parameters for the sampling method, specified by the child sampler.

        Returns:
            :class:`dimod.SampleSet`

        """
        # make sure the BQM is shapeable
        bqm = as_bqm(bqm, cls=[AdjVectorBQM, AdjDictBQM])

        # solve the problem on the child system
        child = self.child
        variables = bqm.variables
        if components is None:
            components = list(connected_components(bqm))
        if isinstance(components, set):
            components = [components]
        sampleset = None
        fixed_value = min(bqm.vartype.value)
        for component in components:
            bqm_copy = bqm.copy()
            bqm_copy.fix_variables({i: fixed_value for i in (variables - component)})
            if sampleset is None:
                # here .truncate(1) is used to pick the best solution only. The other options
                # for future development is to combine all sample with all.
                # This way you'd get the same behaviour as the ExactSolver
                sampleset = child.sample(bqm_copy, **parameters).truncate(1)
            else:
                sampleset = append_variables(sampleset.truncate(1), child.sample(bqm_copy, **parameters).truncate(1))

        if sampleset is None:
            return SampleSet.from_samples_bqm({}, bqm)
        else:
            return SampleSet.from_samples_bqm(sampleset, bqm)

